# -*- encoding: utf-8 -*-

"""
Full Python serializer for Django.
"""

from django.conf import settings
from django.utils.encoding import smart_unicode, smart_str
from django.core.serializers.python import _get_model
from django.db import models
from django.db.models.fields import FieldDoesNotExist

import base


class Serializer(base.Serializer):
    """
    Python serializer for Django modelled after Ruby on Rails.
    Default behaviour is to serialize only model fields with the exception
    of ForeignKey and ManyToMany fields which must be explicitly added in the
    ``relations`` argument.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize instance attributes.
        """
        self._fields = None
        self._extras = None
        self.objects = []
        super(Serializer, self).__init__(*args, **kwargs)

    def start_serialization(self):
        """
        Called when serializing of the queryset starts.
        """
        self._fields = None
        self._extras = None
        self.objects = []

    def end_serialization(self):
        """
        Called when serializing of the queryset ends.
        """
        pass

    def start_object(self, obj):
        """
        Called when serializing of an object starts.
        """
        self._fields = {}
        self._extras = {}
        
    def end_object(self, obj):
        """
        Called when serializing of an object ends.
        """
        self.objects.append({
            "model"  : smart_unicode(obj._meta),
            "fields" : self._fields})
        if self._extras:
            self.objects[-1]["extras"] = self._extras
        if 'id' not in self.excludes and 'pk' not in self.excludes:
            self.objects[-1]["pk"] = smart_unicode(obj._get_pk_val(), strings_only=True)
        self._fields = None
        self._extras = None

    def handle_field(self, obj, field):
        """
        Called to handle each individual (non-relational) field on an object.
        """
        if getattr(field, 'dumb_rel', False):
            self._fields[field.name] = field.value_to_string(obj)
        else:
            self._fields[field.name] = smart_unicode(
                getattr(obj, field.name), strings_only=True)

    def handle_fk_field(self, obj, field):
        """
        Called to handle a ForeignKey field.
        Recursively serializes relations specified in the 'relations' option.
        """
        fname = field.name
        related = getattr(obj, fname)
        if related is not None:
            if fname in self.relations:
                # perform full serialization of FK
                serializer = Serializer()
                options = {}
                if isinstance(self.relations, dict):
                    if isinstance(self.relations[fname], dict):
                        options = self.relations[fname]
                self._fields[fname] = serializer.serialize([related],
                    **options)[0]
            else:
                # emulate the original behaviour and serialize the pk value
                if field.rel.field_name == related._meta.pk.name:
                    # Related to remote object via primary key
                    related = related._get_pk_val()
                else:
                    # Related to remote object via other field
                    related = getattr(related, field.rel.field_name)
                self._fields[fname] = smart_unicode(related, strings_only=True)
        else:
            self._fields[fname] = smart_unicode(related, strings_only=True)

    def handle_m2m_field(self, obj, field):
        """
        Called to handle a ManyToManyField.
        Recursively serializes relations specified in the 'relations' option.
        """
        fname = field.name
        if fname in self.relations:
            # perform full serialization of M2M
            serializer = Serializer()
            options = {}
            if isinstance(self.relations, dict):
                if isinstance(self.relations[fname], dict):
                    options = self.relations[fname]
            self._fields[fname] = [
                serializer.serialize([related], **options)[0]
                for related in getattr(obj, fname).iterator()]
        else:
            # emulate the original behaviour and serialize to a list of 
            # primary key values
            self._fields[fname] = [
                smart_unicode(related._get_pk_val(), strings_only=True)
                   for related in getattr(obj, fname).iterator()]

    def handle_related_field(self, obj, field):
        serializer = Serializer()
        options = {}
        fname = field.get_accessor_name()
        if isinstance(self.relations, dict) and isinstance(self.relations[fname], dict):
            options = self.relations[fname]
        self._fields[fname] = [
            serializer.serialize([related], **options)[0]
            for related in getattr(obj, fname).iterator()]

    def getvalue(self):
        """
        Return the fully serialized queryset (or None if the output stream is
        not seekable).
        """
        return self.objects
    
    def handle_extra_field(self, obj, field):
        """
        Return "extra" fields that the user specifies.
        Can be a property or callable that takes no arguments.
        """
        if hasattr(obj, field):
            extra = getattr(obj, field)
            if callable(extra):
                self._extras[field] = smart_unicode(extra(), strings_only=True)
            else:
                self._extras[field] = smart_unicode(extra, strings_only=True)

    def handle_field_extra_field(self, obj, field_name):
        self._fields[field_name] = smart_unicode(getattr(obj, field_name),
            strings_only=True)
        

def Deserializer(object_list, **options):
    """
    Deserialize simple Python objects back into Django ORM instances.

    It's expected that you pass the Python objects themselves (instead of a
    stream or a string) to the constructor
    """
    models.get_apps()
    for d in object_list:
        # Look up the model and starting build a dict of data for it.
        Model = _get_model(d['model'])
        data = {}
        if 'pk' in d:
            data = {Model._meta.pk.attname : Model._meta.pk.to_python(d['pk'])}
        m2m_data = {}
        related_data = {}
        fk_inline_data = {}
        
        # Handle each field
        for (field_name, field_value) in d["fields"].iteritems():
            if isinstance(field_value, str):
                field_value = smart_unicode(
                    field_value,
                    options.get("encoding", settings.DEFAULT_CHARSET),
                    strings_only=True)

            # Handle related fields
            try:
                field = Model._meta.get_field(field_name)
            except FieldDoesNotExist:
                field = None
                for field in Model._meta.get_all_related_objects():
                    if field.model and field.get_accessor_name() == field_name:
                        break
                if field:
                    related_data[field_name] = list(
                        Deserializer(field_value, **options))
                continue
            
            # Handle M2M relations
            if field.rel and isinstance(field.rel, models.ManyToManyRel):
                m2m_convert = field.rel.to._meta.pk.to_python
                m2m_data[field.name] = [m2m_convert(smart_unicode(pk))
                                        for pk in field_value]

            # Handle FK fields
            elif field.rel and isinstance(field.rel, models.ManyToOneRel):
                # handle inline fk fields
                if type(field_value) in (dict, list):
                    fk_inline_data[field_name] = list(
                        Deserializer([field_value], **options))[0]
                else: # handle normal fk fields
                    if field_value is not None:
                        data[field.attname] = field.rel.to._meta.get_field(
                            field.rel.field_name).to_python(field_value)
                    else:
                        data[field.attname] = None

            # Handle all other fields
            else:
                data[field.name] = field.to_python(field_value)

        yield DeserializedObject(Model(**data), m2m_data, related_data, fk_inline_data)


class DeserializedObject(object):
    """
    A deserialized model.

    Basically a container for holding the pre-saved deserialized data along
    with the many-to-many data saved with the object.

    Call ``save()`` to save the object (with the many-to-many data) to the
    database; call ``save(save_m2m=False)`` to save just the object fields
    (and not touch the many-to-many stuff.)
    """

    def __init__(self, obj, m2m_data = None, related_data = None, fk_inline_data =None):
        self.object = obj
        self.m2m_data = m2m_data
        self.related_data = related_data
        self.fk_inline_data = fk_inline_data

    def __repr__(self):
        return "<DeserializedObject: %s>" % smart_str(self.object)

    def save(self, save_m2m=True, save_related = True, forced_id = None):
        if forced_id is not None:
            self.object.id = forced_id
        if self.fk_inline_data:
            for accessor_name, obj in self.fk_inline_data.iteritems():
                obj.save()
                setattr(self.object, accessor_name, obj.object)
        # здесь не raw = True из-за выставления времен создания версии и автора
        models.Model.save_base(self.object) #, raw = True)
        if self.m2m_data and save_m2m:
            for accessor_name, object_list in self.m2m_data.items():
                setattr(self.object, accessor_name, object_list)
        # prevent a second (possibly accidental) call to save() from saving
        # the m2m data twice.
        self.m2m_data = None
        if self.related_data and save_related:
            for accessor_name, object_list in self.related_data.iteritems():
                for obj in object_list:
                    obj.save()
                setattr(self.object, accessor_name, [obj.object for obj in object_list])

