# -*- encoding: utf-8 -*-

from collections import defaultdict
import datetime
from datetime import timedelta

from django import forms
from django.http import HttpResponseRedirect, HttpResponseNotFound
from django.shortcuts import get_object_or_404
from django.core.urlresolvers import reverse
from django.contrib.auth.decorators import login_required
from annoying.decorators import render_to

from utils import form_maker
from timetable.models import Gym, TTItem, TTFix, Period, weekday_names, week_days
from timetable.forms import TTItemFormset, TTFixFormset, PeriodForm


@login_required
@render_to('timetable/timetable_edit.html')
def timetable_edit(request, id = None, period_id = None):
    ''' Редактирование постоянного расписания и поправок для одного зала '''
    if not id:
        try:
            gym = Gym.objects.all()[:1].get()
        except Gym.DoesNotExist:
            return HttpResponseNotFound()
    else:
        gym = get_object_or_404(Gym, pk = id)
    if 'create' in request.GET:
        return HttpResponseRedirect(reverse('period_create', kwargs = {'id': gym.id}))
    elif 'edit' in request.GET and request.GET.get('period'):
        return HttpResponseRedirect(reverse('period_edit', kwargs = {
            'id': request.GET['period']}))
    elif 'show' in request.GET and request.GET.get('period'):
        return HttpResponseRedirect(reverse('timetable_edit_period', kwargs = {
            'id': gym.id, 'period_id': request.GET['period']}))
    make_form = form_maker(request)
    # основное расписание
    item_period = get_object_or_404(Period, pk = period_id) if period_id else \
                  Period.get_current(gym)
    item_formsets = []
    for d in week_days:
        formset = make_form(
            TTItemFormset, instance = gym, prefix = str(d),
            queryset = gym.ttitem_set.filter(day_of_week = d, period = item_period))
        formset.weekday = weekday_names[d]
        formset.day_of_week = d
        item_formsets.append(formset)
    # форма для периода
    today = datetime.date.today()
    class ChoosePeriodForm(forms.Form):
        period = forms.ModelChoiceField(queryset = gym.periods.filter(
            end__gt = today), initial = item_period.id if item_period else None)
    period_form = make_form(ChoosePeriodForm)
    # поправки к расписанию
    fix_formset = make_form(
        TTFixFormset, instance = gym,
        queryset = gym.ttfix_set.filter(
            date__lte = datetime.date.today() + timedelta(days = 30)))
    fix_formset.forms = list(reversed(fix_formset.forms))
    if request.method == 'POST' and all(fs.is_valid()
                                        for fs in item_formsets + [fix_formset]):
        for fs in item_formsets:
            for form in fs.forms:
                item = form.save(commit = False)
                if item.start and item.end:
                    item.day_of_week = fs.day_of_week
                    item.period = item_period
                    item.save()
                elif item.id:
                    item.delete()
        fix_formset.save()
        return HttpResponseRedirect(
            reverse('timetable_edit_period', kwargs = {
                'id': gym.id, 'period_id': period_id}) if period_id else
            gym.get_timetable_edit_url())
    return {'gym': gym, 'all_gyms': Gym.objects.all(),
            'item_formsets': item_formsets, 'fix_formset': fix_formset,
            'period_form': period_form, 'item_period': item_period}


@login_required
@render_to('timetable/period_edit.html')
def period_create(request, id):
    gym = get_object_or_404(Gym, pk = id)
    form = form_maker(request)(PeriodForm, initial = {'gym': gym.id})
    if request.method == 'POST' and form.is_valid():
        period = form.save(commit = False)
        period.gym = gym
        period.save()
        return HttpResponseRedirect(gym.get_timetable_edit_url())
    return {'gym': gym, 'form': form}


@login_required
@render_to('timetable/period_edit.html')
def period_edit(request, id):
    period = get_object_or_404(Period, pk = id)
    form = form_maker(request)(PeriodForm, instance = period)
    if request.method == 'POST':
        redirect = HttpResponseRedirect(period.gym.get_timetable_edit_url())
        if 'delete' in request.POST:
            period.delete()
            return redirect
        if form.is_valid():
            form.save()
            return redirect
    return {'period': period, 'form': form, 'gym': period.gym}


def get_items_by_gym_by_date(period_start, n_days = 7):
    ''' Словарь, где по залу и дню мы получаем список TTBase '''
    assert n_days <= 7
    def d_by_gym_by_date(items_with_dates):
        items_by_gym_by_date = {}
        for item, date in items_with_dates:
            if item.gym not in items_by_gym_by_date:
                items_by_gym_by_date[item.gym] = defaultdict(list)
            item.date = date
            items_by_gym_by_date[item.gym][item.date].append(item)
        return items_by_gym_by_date
    date_by_day = dict((n, period_start + timedelta(days = n)) for n in xrange(n_days))
    # TODO - filter by period too
    items_by_gym_by_date = d_by_gym_by_date(
        (item, date_by_day[item.day_of_week])
        for item in TTItem.objects.select_related('gym', 'period')
        if item.period.applies(date_by_day[item.day_of_week]))
    fixes_by_gym_by_date = d_by_gym_by_date(
        (fix, fix.date) for fix in TTFix.objects\
        .filter(date__gte = period_start,
                date__lte = period_start + timedelta(days = n_days))\
        .select_related('gym'))
    # мержим: у fix больший приоритет
    for gym, fixes_by_date in fixes_by_gym_by_date.iteritems():
        items_by_date = items_by_gym_by_date.get(gym)
        if not items_by_date: continue
        for date, fixes in fixes_by_date.iteritems():
            items_by_date[date] = fixes
    return items_by_gym_by_date


@render_to('timetable/week_timetable.html')
def week_timetable(request, skip = 0):
    today = datetime.date.today()
    skip = int(skip)
    week_start = today - timedelta(days = today.weekday()) + timedelta(days = 7 * skip)
    dates = [week_start + timedelta(days = n) for n in xrange(7)]
    # вместо словаря по датам будет список в порядке дат
    items_by_gym_by_date = get_items_by_gym_by_date(week_start)
    for gym in items_by_gym_by_date.keys():
        items_by_gym_by_date[gym] = [items_by_gym_by_date[gym][d] for d in dates]
    return {
        'items_by_gym_by_date': sorted(items_by_gym_by_date.iteritems(),
                                       key = lambda (g, v): unicode(g)),
        'dates': dates,
        'next_url': reverse('week_timetable_skip', kwargs = {'skip': skip + 1}),
        'prev_url': reverse('week_timetable_skip', kwargs = {'skip': skip - 1}) \
        if skip else None,
        }


@render_to('timetable/day_timetable.html')
def day_timetable(request, skip = 0):
    ''' Расписание на день - с возможностью листать на неделю вперед без загрузки страницы '''
    skip, n_days = int(skip), 7
    period_start = datetime.date.today() + timedelta(days = n_days * skip)
    items_by_gym_by_date = get_items_by_gym_by_date(period_start)
    # перегруппируем по датам
    items_by_date_by_gym = defaultdict(dict)
    for gym, items_by_date in items_by_gym_by_date.iteritems():
        for date, items in items_by_date.iteritems():
            items_by_date_by_gym[date][gym] = items
    for date in items_by_date_by_gym.keys():
        for gym in items_by_gym_by_date.iterkeys():
            if gym not in items_by_date_by_gym[date]:
                items_by_date_by_gym[date][gym] = []
        items_by_date_by_gym[date] = sorted(items_by_date_by_gym[date].iteritems(),
                                            key = lambda (k, v): unicode(k))
    url_for_skip = lambda s: reverse('day_timetable_skip', kwargs = {'skip': s})
    return {
        'items_by_date_by_gym': sorted(items_by_date_by_gym.iteritems(),
                                       key = lambda (k, v): k),
        'dates': [period_start + timedelta(days = n) for n in xrange(n_days)],
        'next_url': url_for_skip(skip + 1),
        'prev_url': url_for_skip(skip - 1) if skip else None,
        'cur_url': url_for_skip(skip),
        'shown_index': int(request.GET.get('show_index', 0)),
        'last_index': n_days - 1,
        }
